@@ -1,4 +1,4 @@
-# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -11,23 +11,16 @@ r"""
     :connectstring: oracle+cx_oracle://user:pass@host:port/dbname[?key=value&key=value...]
     :url: https://oracle.github.io/python-cx_Oracle/
 
-DSN vs. Hostname connections
------------------------------
+Additional Connect Arguments
+----------------------------
 
-The dialect will connect to a DSN if no database name portion is presented,
-such as::
+When connecting with the ``dbname`` URL token present, the ``hostname``,
+``port``, and ``dbname`` tokens are converted to a TNS name using the
+``cx_Oracle.makedsn()`` function. The URL below::
 
-    engine = create_engine("oracle+cx_oracle://scott:tiger@oracle1120/?encoding=UTF-8&nencoding=UTF-8")
+    e = create_engine("oracle+cx_oracle://user:pass@hostname/dbname")
 
-Above, ``oracle1120`` is passed to cx_Oracle as an Oracle datasource name.
-
-Alternatively, if a database name is present, the ``cx_Oracle.makedsn()``
-function is used to create an ad-hoc "datasource" name assuming host
-and port::
-
-    engine = create_engine("oracle+cx_oracle://scott:tiger@hostname:1521/dbname?encoding=UTF-8&nencoding=UTF-8")
-
-Above, the DSN would be created as follows::
+Will be used to create the DSN as follows::
 
     >>> import cx_Oracle
     >>> cx_Oracle.makedsn("hostname", 1521, sid="dbname")
@@ -36,44 +29,38 @@ Above, the DSN would be created as follows::
 The ``service_name`` parameter, also consumed by ``cx_Oracle.makedsn()``, may
 be specified in the URL query string, e.g. ``?service_name=my_service``.
 
+If ``dbname`` is not present, then the value of ``hostname`` in the
+URL is used directly as the DSN passed to ``cx_Oracle.connect()``.
 
-Passing cx_Oracle connect arguments
------------------------------------
-
-Additional connection arguments can usually be passed via the URL
-query string; particular symbols like ``cx_Oracle.SYSDBA`` are intercepted
-and converted to the correct symbol::
-
-    e = create_engine(
-        "oracle+cx_oracle://user:pass@dsn?encoding=UTF-8&nencoding=UTF-8&mode=SYSDBA&events=true")
-
-.. versionchanged:: 1.3 the cx_oracle dialect now accepts all argument names
-   within the URL string itself, to be passed to the cx_Oracle DBAPI.   As
-   was the case earlier but not correctly documented, the
-   :paramref:`_sa.create_engine.connect_args` parameter also accepts all
-   cx_Oracle DBAPI connect arguments.
-
-To pass arguments directly to ``.connect()`` wihtout using the query
-string, use the :paramref:`_sa.create_engine.connect_args` dictionary.
+Additional connection arguments may be sent to the ``cx_Oracle.connect()``
+function using the :paramref:`.create_engine.connect_args` dictionary.
 Any cx_Oracle parameter value and/or constant may be passed, such as::
 
     import cx_Oracle
     e = create_engine(
         "oracle+cx_oracle://user:pass@dsn",
         connect_args={
-            "encoding": "UTF-8",
-            "nencoding": "UTF-8",
             "mode": cx_Oracle.SYSDBA,
             "events": True
         }
     )
 
-Options consumed by the SQLAlchemy cx_Oracle dialect outside of the driver
---------------------------------------------------------------------------
+Alternatively, most cx_Oracle DBAPI arguments can also be encoded as strings
+within the URL, which includes parameters such as ``mode``, ``purity``,
+``events``, ``threaded``, and others::
+
+    e = create_engine(
+        "oracle+cx_oracle://user:pass@dsn?mode=SYSDBA&events=true")
+
+.. versionchanged:: 1.3 the cx_oracle dialect now accepts all argument names
+   within the URL string itself, to be passed to the cx_Oracle DBAPI.   As
+   was the case earlier but not correctly documented, the
+   :paramref:`.create_engine.connect_args` parameter also accepts all
+   cx_Oracle DBAPI connect arguments.
 
 There are also options that are consumed by the SQLAlchemy cx_oracle dialect
-itself.  These options are always passed directly to :func:`_sa.create_engine`
-, such as::
+itself.  These options are always passed directly to :func:`.create_engine`,
+such as::
 
     e = create_engine(
         "oracle+cx_oracle://user:pass@dsn", coerce_to_unicode=False)
@@ -98,63 +85,28 @@ The parameters accepted by the cx_oracle dialect are as follows:
 Unicode
 -------
 
-As is the case for all DBAPIs under Python 3, all strings are inherently
-Unicode strings.     Under Python 2, cx_Oracle also supports Python Unicode
-objects directly.    In all cases however, the driver requires an explicit
-encoding configuration.
+The cx_Oracle DBAPI as of version 5 fully supports Unicode, and has the
+ability to return string results as Python Unicode objects natively.
 
-Ensuring the Correct Client Encoding
-^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-
-The long accepted standard for establishing client encoding for nearly all
-Oracle related software is via the `NLS_LANG <https://www.oracle.com/database/technologies/faq-nls-lang.html>`_
-environment variable.   cx_Oracle like most other Oracle drivers will use
-this environment variable as the source of its encoding configuration.  The
-format of this variable is idiosyncratic; a typical value would be
-``AMERICAN_AMERICA.AL32UTF8``.
-
-The cx_Oracle driver also supports a programmatic alternative which is to
-pass the ``encoding`` and ``nencoding`` parameters directly to its
-``.connect()`` function.  These can be present in the URL as follows::
-
-    engine = create_engine("oracle+cx_oracle://scott:tiger@oracle1120/?encoding=UTF-8&nencoding=UTF-8")
-
-For the meaning of the ``encoding`` and ``nencoding`` parameters, please
-consult
-`Characters Sets and National Language Support (NLS) <https://cx-oracle.readthedocs.io/en/latest/user_guide/globalization.html#globalization>`_.
-
-.. seealso::
-
-    `Characters Sets and National Language Support (NLS) <https://cx-oracle.readthedocs.io/en/latest/user_guide/globalization.html#globalization>`_
-    - in the cx_Oracle documentation.
-
-
-Unicode-specific Column datatypes
-^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-
-The Core expression language handles unicode data by use of the :class:`.Unicode`
-and :class:`.UnicodeText`
-datatypes.  These types correspond to the  VARCHAR2 and CLOB Oracle datatypes by
+Explicit Unicode support is available by using the :class:`.Unicode` datatype
+with SQLAlchemy Core expression language, as well as the :class:`.UnicodeText`
+datatype.  These types correspond to the  VARCHAR2 and CLOB Oracle datatypes by
 default.   When using these datatypes with Unicode data, it is expected that
 the Oracle database is configured with a Unicode-aware character set, as well
 as that the ``NLS_LANG`` environment variable is set appropriately, so that
 the VARCHAR2 and CLOB datatypes can accommodate the data.
 
 In the case that the Oracle database is not configured with a Unicode character
-set, the two options are to use the :class:`_types.NCHAR` and
-:class:`_oracle.NCLOB` datatypes explicitly, or to pass the flag
-``use_nchar_for_unicode=True`` to :func:`_sa.create_engine`,
-which will cause the
+set, the two options are to use the :class:`.oracle.NCHAR` and
+:class:`.oracle.NCLOB` datatypes explicitly, or to pass the flag
+``use_nchar_for_unicode=True`` to :func:`.create_engine`, which will cause the
 SQLAlchemy dialect to use NCHAR/NCLOB for the :class:`.Unicode` /
 :class:`.UnicodeText` datatypes instead of VARCHAR/CLOB.
 
 .. versionchanged:: 1.3  The :class:`.Unicode` and :class:`.UnicodeText`
    datatypes now correspond to the ``VARCHAR2`` and ``CLOB`` Oracle datatypes
    unless the ``use_nchar_for_unicode=True`` is passed to the dialect
-   when :func:`_sa.create_engine` is called.
-
-Unicode Coercion of result rows under Python 2
-^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+   when :func:`.create_engine` is called.
 
 When result sets are fetched that include strings, under Python 3 the cx_Oracle
 DBAPI returns all strings as Python Unicode objects, since Python 3 only has a
@@ -166,13 +118,14 @@ made use of converters that were supplied by cx_Oracle but were found to be
 non-performant; SQLAlchemy's own converters are used for the string to Unicode
 conversion under Python 2.  To disable the Python 2 Unicode conversion for
 VARCHAR2, CHAR, and CLOB, the flag ``coerce_to_unicode=False`` can be passed to
-:func:`_sa.create_engine`.
+:func:`.create_engine`.
 
 .. versionchanged:: 1.3 Unicode conversion is applied to all string values
    by default under python 2.  The ``coerce_to_unicode`` now defaults to True
    and can be set to False to disable the Unicode coercion of strings that are
    delivered as VARCHAR2/CHAR/CLOB data.
 
+
 .. _cx_oracle_unicode_encoding_errors:
 
 Encoding Errors
@@ -207,7 +160,7 @@ altering the type coercion behavior at the same time.
 
 Users of the cx_Oracle dialect are **strongly encouraged** to read through
 cx_Oracle's list of built-in datatype symbols at
-http://cx-oracle.readthedocs.io/en/latest/module.html#database-types.
+http://cx-oracle.readthedocs.io/en/latest/module.html#types.
 Note that in some cases, significant performance degradation can occur when
 using these types vs. not, in particular when specifying ``cx_Oracle.CLOB``.
 
@@ -282,7 +235,7 @@ As of the 6 series, this limitation has been lifted.  Nevertheless, because
 SQLAlchemy pre-reads these LOBs up front, this issue is avoided in any case.
 
 To disable the auto "read()" feature of the dialect, the flag
-``auto_convert_lobs=False`` may be passed to :func:`_sa.create_engine`.  Under
+``auto_convert_lobs=False`` may be passed to :func:`.create_engine`.  Under
 the cx_Oracle 5 series, having this flag turned off means there is the chance
 of reading from a stale LOB object if not read as it is fetched.   With
 cx_Oracle 6, this issue is resolved.
@@ -307,12 +260,12 @@ Precision Numerics
 
 SQLAlchemy's numeric types can handle receiving and returning values as Python
 ``Decimal`` objects or float objects.  When a :class:`.Numeric` object, or a
-subclass such as :class:`.Float`, :class:`_oracle.DOUBLE_PRECISION` etc. is in
+subclass such as :class:`.Float`, :class:`.oracle.DOUBLE_PRECISION` etc. is in
 use, the :paramref:`.Numeric.asdecimal` flag determines if values should be
 coerced to ``Decimal`` upon return, or returned as float objects.   To make
 matters more complicated under Oracle, Oracle's ``NUMBER`` type can also
 represent integer values if the "scale" is zero, so the Oracle-specific
-:class:`_oracle.NUMBER` type takes this into account as well.
+:class:`.oracle.NUMBER` type takes this into account as well.
 
 The cx_Oracle dialect makes extensive use of connection- and cursor-level
 "outputtypehandler" callables in order to coerce numeric values as requested.
@@ -328,7 +281,7 @@ When no typing objects are present, as when executing plain SQL strings, a
 default "outputtypehandler" is present which will generally return numeric
 values which specify precision and scale as Python ``Decimal`` objects.  To
 disable this coercion to decimal for performance reasons, pass the flag
-``coerce_to_decimal=False`` to :func:`_sa.create_engine`::
+``coerce_to_decimal=False`` to :func:`.create_engine`::
 
     engine = create_engine("oracle+cx_oracle://dsn", coerce_to_decimal=False)
 
@@ -514,7 +467,7 @@ class _OracleUnicodeStringNCHAR(oracle.NVARCHAR2):
 
 class _OracleUnicodeStringCHAR(sqltypes.Unicode):
     def get_dbapi_type(self, dbapi):
-        return dbapi.LONG_STRING
+        return None
 
 
 class _OracleUnicodeTextNCLOB(oracle.NCLOB):
@@ -624,63 +577,19 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
                 if bindparam.isoutparam:
                     name = self.compiled.bind_names[bindparam]
                     type_impl = bindparam.type.dialect_impl(self.dialect)
-
                     if hasattr(type_impl, "_cx_oracle_var"):
                         self.out_parameters[name] = type_impl._cx_oracle_var(
                             self.dialect, self.cursor
                         )
                     else:
                         dbtype = type_impl.get_dbapi_type(self.dialect.dbapi)
-
-                        cx_Oracle = self.dialect.dbapi
-
                         if dbtype is None:
                             raise exc.InvalidRequestError(
-                                "Cannot create out parameter for "
-                                "parameter "
+                                "Cannot create out parameter for parameter "
                                 "%r - its type %r is not supported by"
                                 " cx_oracle" % (bindparam.key, bindparam.type)
                             )
-
-                        if compat.py2k and dbtype in (
-                            cx_Oracle.CLOB,
-                            cx_Oracle.NCLOB,
-                        ):
-                            outconverter = (
-                                processors.to_unicode_processor_factory(
-                                    self.dialect.encoding,
-                                    errors=self.dialect.encoding_errors,
-                                )
-                            )
-                            self.out_parameters[name] = self.cursor.var(
-                                dbtype,
-                                outconverter=lambda value: outconverter(
-                                    value.read()
-                                ),
-                            )
-
-                        elif dbtype in (
-                            cx_Oracle.BLOB,
-                            cx_Oracle.CLOB,
-                            cx_Oracle.NCLOB,
-                        ):
-                            self.out_parameters[name] = self.cursor.var(
-                                dbtype, outconverter=lambda value: value.read()
-                            )
-                        elif compat.py2k and isinstance(
-                            type_impl, sqltypes.Unicode
-                        ):
-                            outconverter = (
-                                processors.to_unicode_processor_factory(
-                                    self.dialect.encoding,
-                                    errors=self.dialect.encoding_errors,
-                                )
-                            )
-                            self.out_parameters[name] = self.cursor.var(
-                                dbtype, outconverter=outconverter
-                            )
-                        else:
-                            self.out_parameters[name] = self.cursor.var(dbtype)
+                        self.out_parameters[name] = self.cursor.var(dbtype)
                     self.parameters[0][
                         quoted_bind_names.get(name, name)
                     ] = self.out_parameters[name]
@@ -861,7 +770,7 @@ class OracleDialect_cx_oracle(OracleDialect):
             "in a future release.  As of version 1.3, it defaults to False "
             "rather than True.  The 'threaded' option can be passed to "
             "cx_Oracle directly in the URL query string passed to "
-            ":func:`_sa.create_engine`.",
+            ":func:`.create_engine`.",
         )
     )
     def __init__(
@@ -968,66 +877,6 @@ class OracleDialect_cx_oracle(OracleDialect):
 
         self._detect_decimal_char(connection)
 
-    def get_isolation_level(self, connection):
-        # sources:
-
-        # general idea of transaction id, have to start one, etc.
-        # https://stackoverflow.com/questions/10711204/how-to-check-isoloation-level
-
-        # how to decode xid cols from v$transaction to match
-        # https://asktom.oracle.com/pls/apex/f?p=100:11:0::::P11_QUESTION_ID:9532779900346079444
-
-        # Oracle tuple comparison without using IN:
-        # https://www.sql-workbench.eu/comparison/tuple_comparison.html
-
-        with connection.cursor() as cursor:
-            # this is the only way to ensure a transaction is started without
-            # actually running DML.   There's no way to see the configured
-            # isolation level without getting it from v$transaction which
-            # means transaction has to be started.
-            outval = cursor.var(str)
-            cursor.execute(
-                """
-                begin
-                   :trans_id := dbms_transaction.local_transaction_id( TRUE );
-                end;
-                """,
-                {"trans_id": outval},
-            )
-            trans_id = outval.getvalue()
-            xidusn, xidslot, xidsqn = trans_id.split(".", 2)
-
-            cursor.execute(
-                "SELECT CASE BITAND(t.flag, POWER(2, 28)) "
-                "WHEN 0 THEN 'READ COMMITTED' "
-                "ELSE 'SERIALIZABLE' END AS isolation_level "
-                "FROM v$transaction t WHERE "
-                "(t.xidusn, t.xidslot, t.xidsqn) = "
-                "((:xidusn, :xidslot, :xidsqn))",
-                {"xidusn": xidusn, "xidslot": xidslot, "xidsqn": xidsqn},
-            )
-            row = cursor.fetchone()
-            if row is None:
-                raise exc.InvalidRequestError(
-                    "could not retrieve isolation level"
-                )
-            result = row[0]
-
-        return result
-
-    def set_isolation_level(self, connection, level):
-        if hasattr(connection, "connection"):
-            dbapi_connection = connection.connection
-        else:
-            dbapi_connection = connection
-        if level == "AUTOCOMMIT":
-            dbapi_connection.autocommit = True
-        else:
-            dbapi_connection.autocommit = False
-            connection.rollback()
-            with connection.cursor() as cursor:
-                cursor.execute("ALTER SESSION SET ISOLATION_LEVEL=%s" % level)
-
     def _detect_decimal_char(self, connection):
         # we have the option to change this setting upon connect,
         # or just look at what it is upon connect and convert.
@@ -1077,11 +926,7 @@ class OracleDialect_cx_oracle(OracleDialect):
         def output_type_handler(
             cursor, name, default_type, size, precision, scale
         ):
-
-            if (
-                default_type == cx_Oracle.NUMBER
-                and default_type is not cx_Oracle.NATIVE_FLOAT
-            ):
+            if default_type == cx_Oracle.NUMBER:
                 if not dialect.coerce_to_decimal:
                     return None
                 elif precision == 0 and scale in (0, -127):
@@ -1103,15 +948,9 @@ class OracleDialect_cx_oracle(OracleDialect):
                     )
 
             # allow all strings to come back natively as Unicode
-            elif (
-                dialect.coerce_to_unicode
-                and default_type
-                in (
-                    cx_Oracle.STRING,
-                    cx_Oracle.FIXED_CHAR,
-                )
-                and default_type is not cx_Oracle.CLOB
-                and default_type is not cx_Oracle.NCLOB
+            elif dialect.coerce_to_unicode and default_type in (
+                cx_Oracle.STRING,
+                cx_Oracle.FIXED_CHAR,
             ):
                 if compat.py2k:
                     outconverter = processors.to_unicode_processor_factory(
@@ -1140,16 +979,17 @@ class OracleDialect_cx_oracle(OracleDialect):
                         dialect.encoding, errors=dialect.encoding_errors
                     )
                     return cursor.var(
-                        cx_Oracle.LONG_STRING,
+                        default_type,
                         size,
                         cursor.arraysize,
-                        outconverter=outconverter,
+                        outconverter=lambda value: outconverter(value.read()),
                     )
                 else:
                     return cursor.var(
-                        cx_Oracle.LONG_STRING,
+                        default_type,
                         size,
                         cursor.arraysize,
+                        outconverter=lambda value: value.read(),
                         **dialect._cursor_var_unicode_kwargs
                     )
 
@@ -1157,9 +997,10 @@ class OracleDialect_cx_oracle(OracleDialect):
                 cx_Oracle.BLOB,
             ):
                 return cursor.var(
-                    cx_Oracle.LONG_BINARY,
+                    default_type,
                     size,
                     cursor.arraysize,
+                    outconverter=lambda value: value.read(),
                 )
 
         return output_type_handler
@@ -1237,13 +1078,14 @@ class OracleDialect_cx_oracle(OracleDialect):
         util.coerce_kw_type(opts, "threaded", bool)
         util.coerce_kw_type(opts, "events", bool)
         util.coerce_kw_type(opts, "purity", convert_cx_oracle_constant)
+
         return ([], opts)
 
     def _get_server_version_info(self, connection):
         return tuple(int(x) for x in connection.connection.version.split("."))
 
     def is_disconnect(self, e, connection, cursor):
-        (error,) = e.args
+        error, = e.args
         if isinstance(
             e, (self.dbapi.InterfaceError, self.dbapi.DatabaseError)
         ) and "not connected" in str(e):
