@@ -1,5 +1,5 @@
 # sql/functions.py
-# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -8,19 +8,41 @@
 """SQL function API, factories, and built-in functions.
 
 """
-from . import sqltypes, schema
-from .base import Executable, ColumnCollection
-from .elements import ClauseList, Cast, Extract, _literal_as_binds, \
-    literal_column, _type_from_args, ColumnElement, _clone,\
-    Over, BindParameter, FunctionFilter, Grouping, WithinGroup
-from .selectable import FromClause, Select, Alias
-from . import util as sqlutil
+from . import annotation
 from . import operators
+from . import schema
+from . import sqltypes
+from . import util as sqlutil
+from .base import ColumnCollection
+from .base import Executable
+from .elements import _clone
+from .elements import _literal_as_binds
+from .elements import _type_from_args
+from .elements import BinaryExpression
+from .elements import BindParameter
+from .elements import Cast
+from .elements import ClauseList
+from .elements import ColumnElement
+from .elements import Extract
+from .elements import FunctionFilter
+from .elements import Grouping
+from .elements import literal_column
+from .elements import Over
+from .elements import WithinGroup
+from .selectable import Alias
+from .selectable import FromClause
+from .selectable import Select
 from .visitors import VisitableType
 from .. import util
-from . import annotation
+
 
 _registry = util.defaultdict(dict)
+_case_sensitive_registry = util.defaultdict(lambda: util.defaultdict(dict))
+_CASE_SENSITIVE = util.symbol(
+    name="case_sensitive_function",
+    doc="Symbol to mark the functions that are switched into case-sensitive "
+    "mode.",
+)
 
 
 def register_function(identifier, fn, package="_default"):
@@ -33,7 +55,60 @@ def register_function(identifier, fn, package="_default"):
 
     """
     reg = _registry[package]
-    reg[identifier] = fn
+    case_sensitive_reg = _case_sensitive_registry[package]
+    raw_identifier = identifier
+    identifier = util.text_type(identifier).lower()
+
+    # Check if a function with the same lowercase identifier is registered.
+    if identifier in reg and reg[identifier] is not _CASE_SENSITIVE:
+        if raw_identifier in case_sensitive_reg[identifier]:
+            util.warn(
+                "The GenericFunction '{}' is already registered and "
+                "is going to be overriden.".format(identifier)
+            )
+            reg[identifier] = fn
+        else:
+            # If a function with the same lowercase identifier is registered,
+            # then these 2 functions are considered as case-sensitive.
+            # Note: This case should raise an error in a later release.
+            util.warn_deprecated(
+                "GenericFunction '{}' is already registered with "
+                "different letter case, so the previously registered function "
+                "'{}' is switched into case-sensitive mode. "
+                "GenericFunction objects will be fully case-insensitive in a "
+                "future release.".format(
+                    raw_identifier,
+                    list(case_sensitive_reg[identifier].keys())[0],
+                )
+            )
+            reg[identifier] = _CASE_SENSITIVE
+
+    # Check if a function with different letter case identifier is registered.
+    elif identifier in case_sensitive_reg:
+        # Note: This case will be removed in a later release.
+        if raw_identifier not in case_sensitive_reg[identifier]:
+            util.warn_deprecated(
+                "GenericFunction(s) '{}' are already registered with "
+                "different letter cases and might interact with '{}'. "
+                "GenericFunction objects will be fully case-insensitive in a "
+                "future release.".format(
+                    sorted(case_sensitive_reg[identifier].keys()),
+                    raw_identifier,
+                )
+            )
+
+        else:
+            util.warn(
+                "The GenericFunction '{}' is already registered and "
+                "is going to be overriden.".format(raw_identifier)
+            )
+
+    # Register by default
+    else:
+        reg[identifier] = fn
+
+    # Always register in case-sensitive registry
+    case_sensitive_reg[identifier][raw_identifier] = fn
 
 
 class FunctionElement(Executable, ColumnElement, FromClause):
@@ -41,6 +116,8 @@ class FunctionElement(Executable, ColumnElement, FromClause):
 
     .. seealso::
 
+        :ref:`coretutorial_functions` - in the Core tutorial
+
         :class:`.Function` - named SQL function.
 
         :data:`.func` - namespace which produces registered or ad-hoc
@@ -53,14 +130,29 @@ class FunctionElement(Executable, ColumnElement, FromClause):
 
     packagenames = ()
 
+    _has_args = False
+
     def __init__(self, *clauses, **kwargs):
-        """Construct a :class:`.FunctionElement`.
+        r"""Construct a :class:`.FunctionElement`.
+
+        :param \*clauses: list of column expressions that form the arguments
+         of the SQL function call.
+
+        :param \**kwargs:  additional kwargs are typically consumed by
+         subclasses.
+
+        .. seealso::
+
+            :data:`.func`
+
+            :class:`.Function`
+
         """
         args = [_literal_as_binds(c, self.name) for c in clauses]
+        self._has_args = self._has_args or bool(args)
         self.clause_expr = ClauseList(
-            operator=operators.comma_op,
-            group_contents=True, *args).\
-            self_group()
+            operator=operators.comma_op, group_contents=True, *args
+        ).self_group()
 
     def _execute_on_connection(self, connection, multiparams, params):
         return connection._execute_function(self, multiparams, params)
@@ -111,15 +203,13 @@ class FunctionElement(Executable, ColumnElement, FromClause):
 
         See :func:`~.expression.over` for a full description.
 
-        .. versionadded:: 0.7
-
         """
         return Over(
             self,
             partition_by=partition_by,
             order_by=order_by,
             rows=rows,
-            range_=range_
+            range_=range_,
         )
 
     def within_group(self, *order_by):
@@ -166,12 +256,77 @@ class FunctionElement(Executable, ColumnElement, FromClause):
             return self
         return FunctionFilter(self, *criterion)
 
+    def as_comparison(self, left_index, right_index):
+        """Interpret this expression as a boolean comparison between two values.
+
+        A hypothetical SQL function "is_equal()" which compares to values
+        for equality would be written in the Core expression language as::
+
+            expr = func.is_equal("a", "b")
+
+        If "is_equal()" above is comparing "a" and "b" for equality, the
+        :meth:`.FunctionElement.as_comparison` method would be invoked as::
+
+            expr = func.is_equal("a", "b").as_comparison(1, 2)
+
+        Where above, the integer value "1" refers to the first argument of the
+        "is_equal()" function and the integer value "2" refers to the second.
+
+        This would create a :class:`.BinaryExpression` that is equivalent to::
+
+            BinaryExpression("a", "b", operator=op.eq)
+
+        However, at the SQL level it would still render as
+        "is_equal('a', 'b')".
+
+        The ORM, when it loads a related object or collection, needs to be able
+        to manipulate the "left" and "right" sides of the ON clause of a JOIN
+        expression. The purpose of this method is to provide a SQL function
+        construct that can also supply this information to the ORM, when used
+        with the :paramref:`.relationship.primaryjoin` parameter.  The return
+        value is a containment object called :class:`.FunctionAsBinary`.
+
+        An ORM example is as follows::
+
+            class Venue(Base):
+                __tablename__ = 'venue'
+                id = Column(Integer, primary_key=True)
+                name = Column(String)
+
+                descendants = relationship(
+                    "Venue",
+                    primaryjoin=func.instr(
+                        remote(foreign(name)), name + "/"
+                    ).as_comparison(1, 2) == 1,
+                    viewonly=True,
+                    order_by=name
+                )
+
+        Above, the "Venue" class can load descendant "Venue" objects by
+        determining if the name of the parent Venue is contained within the
+        start of the hypothetical descendant value's name, e.g. "parent1" would
+        match up to "parent1/child1", but not to "parent2/child1".
+
+        Possible use cases include the "materialized path" example given above,
+        as well as making use of special SQL functions such as geometric
+        functions to create join conditions.
+
+        :param left_index: the integer 1-based index of the function argument
+         that serves as the "left" side of the expression.
+        :param right_index: the integer 1-based index of the function argument
+         that serves as the "right" side of the expression.
+
+        .. versionadded:: 1.3
+
+        """
+        return FunctionAsBinary(self, left_index, right_index)
+
     @property
     def _from_objects(self):
         return self.clauses._from_objects
 
     def get_children(self, **kwargs):
-        return self.clause_expr,
+        return (self.clause_expr,)
 
     def _copy_internals(self, clone=_clone, **kw):
         self.clause_expr = clone(self.clause_expr, **kw)
@@ -220,7 +375,7 @@ class FunctionElement(Executable, ColumnElement, FromClause):
 
         """
 
-        return Alias(self, name)
+        return Alias._construct(self, name)
 
     def select(self):
         """Produce a :func:`~.expression.select` construct
@@ -265,24 +420,141 @@ class FunctionElement(Executable, ColumnElement, FromClause):
         return self.select().execute()
 
     def _bind_param(self, operator, obj, type_=None):
-        return BindParameter(None, obj, _compared_to_operator=operator,
-                             _compared_to_type=self.type, unique=True,
-                             type_=type_)
+        return BindParameter(
+            None,
+            obj,
+            _compared_to_operator=operator,
+            _compared_to_type=self.type,
+            unique=True,
+            type_=type_,
+        )
 
     def self_group(self, against=None):
         # for the moment, we are parenthesizing all array-returning
         # expressions against getitem.  This may need to be made
         # more portable if in the future we support other DBs
         # besides postgresql.
-        if against is operators.getitem and \
-                isinstance(self.type, sqltypes.ARRAY):
+        if against is operators.getitem and isinstance(
+            self.type, sqltypes.ARRAY
+        ):
             return Grouping(self)
         else:
             return super(FunctionElement, self).self_group(against=against)
 
 
+class FunctionAsBinary(BinaryExpression):
+    def __init__(self, fn, left_index, right_index):
+        left = fn.clauses.clauses[left_index - 1]
+        right = fn.clauses.clauses[right_index - 1]
+
+        self.sql_function = fn
+        self.left_index = left_index
+        self.right_index = right_index
+
+        super(FunctionAsBinary, self).__init__(
+            left,
+            right,
+            operators.function_as_comparison_op,
+            type_=sqltypes.BOOLEANTYPE,
+        )
+
+    @property
+    def left(self):
+        return self.sql_function.clauses.clauses[self.left_index - 1]
+
+    @left.setter
+    def left(self, value):
+        self.sql_function.clauses.clauses[self.left_index - 1] = value
+
+    @property
+    def right(self):
+        return self.sql_function.clauses.clauses[self.right_index - 1]
+
+    @right.setter
+    def right(self, value):
+        self.sql_function.clauses.clauses[self.right_index - 1] = value
+
+    def _copy_internals(self, **kw):
+        clone = kw.pop("clone")
+        self.sql_function = clone(self.sql_function, **kw)
+        super(FunctionAsBinary, self)._copy_internals(**kw)
+
+
 class _FunctionGenerator(object):
-    """Generate :class:`.Function` objects based on getattr calls."""
+    """Generate SQL function expressions.
+
+    :data:`.func` is a special object instance which generates SQL
+    functions based on name-based attributes, e.g.::
+
+        >>> print(func.count(1))
+        count(:param_1)
+
+    The returned object is an instance of :class:`.Function`, and  is a
+    column-oriented SQL element like any other, and is used in that way::
+
+        >>> print(select([func.count(table.c.id)]))
+        SELECT count(sometable.id) FROM sometable
+
+    Any name can be given to :data:`.func`. If the function name is unknown to
+    SQLAlchemy, it will be rendered exactly as is. For common SQL functions
+    which SQLAlchemy is aware of, the name may be interpreted as a *generic
+    function* which will be compiled appropriately to the target database::
+
+        >>> print(func.current_timestamp())
+        CURRENT_TIMESTAMP
+
+    To call functions which are present in dot-separated packages,
+    specify them in the same manner::
+
+        >>> print(func.stats.yield_curve(5, 10))
+        stats.yield_curve(:yield_curve_1, :yield_curve_2)
+
+    SQLAlchemy can be made aware of the return type of functions to enable
+    type-specific lexical and result-based behavior. For example, to ensure
+    that a string-based function returns a Unicode value and is similarly
+    treated as a string in expressions, specify
+    :class:`~sqlalchemy.types.Unicode` as the type:
+
+        >>> print(func.my_string(u'hi', type_=Unicode) + ' ' +
+        ...       func.my_string(u'there', type_=Unicode))
+        my_string(:my_string_1) || :my_string_2 || my_string(:my_string_3)
+
+    The object returned by a :data:`.func` call is usually an instance of
+    :class:`.Function`.
+    This object meets the "column" interface, including comparison and labeling
+    functions.  The object can also be passed the :meth:`~.Connectable.execute`
+    method of a :class:`.Connection` or :class:`.Engine`, where it will be
+    wrapped inside of a SELECT statement first::
+
+        print(connection.execute(func.current_timestamp()).scalar())
+
+    In a few exception cases, the :data:`.func` accessor
+    will redirect a name to a built-in expression such as :func:`.cast`
+    or :func:`.extract`, as these names have well-known meaning
+    but are not exactly the same as "functions" from a SQLAlchemy
+    perspective.
+
+    Functions which are interpreted as "generic" functions know how to
+    calculate their return type automatically. For a listing of known generic
+    functions, see :ref:`generic_functions`.
+
+    .. note::
+
+        The :data:`.func` construct has only limited support for calling
+        standalone "stored procedures", especially those with special
+        parameterization concerns.
+
+        See the section :ref:`stored_procedures` for details on how to use
+        the DBAPI-level ``callproc()`` method for fully traditional stored
+        procedures.
+
+    .. seealso::
+
+        :ref:`coretutorial_functions` - in the Core Tutorial
+
+        :class:`.Function`
+
+    """
 
     def __init__(self, **opts):
         self.__names = []
@@ -290,13 +562,13 @@ class _FunctionGenerator(object):
 
     def __getattr__(self, name):
         # passthru __ attributes; fixes pydoc
-        if name.startswith('__'):
+        if name.startswith("__"):
             try:
                 return self.__dict__[name]
             except KeyError:
                 raise AttributeError(name)
 
-        elif name.endswith('_'):
+        elif name.endswith("_"):
             name = name[0:-1]
         f = _FunctionGenerator(**self.opts)
         f.__names = list(self.__names) + [name]
@@ -316,99 +588,51 @@ class _FunctionGenerator(object):
             package = None
 
         if package is not None:
-            func = _registry[package].get(fname)
+            func = _registry[package].get(fname.lower())
+            if func is _CASE_SENSITIVE:
+                case_sensitive_reg = _case_sensitive_registry[package]
+                func = case_sensitive_reg.get(fname.lower()).get(fname)
+
             if func is not None:
                 return func(*c, **o)
 
-        return Function(self.__names[-1],
-                        packagenames=self.__names[0:-1], *c, **o)
+        return Function(
+            self.__names[-1], packagenames=self.__names[0:-1], *c, **o
+        )
 
 
 func = _FunctionGenerator()
-"""Generate SQL function expressions.
-
-   :data:`.func` is a special object instance which generates SQL
-   functions based on name-based attributes, e.g.::
-
-        >>> print(func.count(1))
-        count(:param_1)
-
-   The element is a column-oriented SQL element like any other, and is
-   used in that way::
-
-        >>> print(select([func.count(table.c.id)]))
-        SELECT count(sometable.id) FROM sometable
-
-   Any name can be given to :data:`.func`. If the function name is unknown to
-   SQLAlchemy, it will be rendered exactly as is. For common SQL functions
-   which SQLAlchemy is aware of, the name may be interpreted as a *generic
-   function* which will be compiled appropriately to the target database::
-
-        >>> print(func.current_timestamp())
-        CURRENT_TIMESTAMP
-
-   To call functions which are present in dot-separated packages,
-   specify them in the same manner::
-
-        >>> print(func.stats.yield_curve(5, 10))
-        stats.yield_curve(:yield_curve_1, :yield_curve_2)
-
-   SQLAlchemy can be made aware of the return type of functions to enable
-   type-specific lexical and result-based behavior. For example, to ensure
-   that a string-based function returns a Unicode value and is similarly
-   treated as a string in expressions, specify
-   :class:`~sqlalchemy.types.Unicode` as the type:
-
-        >>> print(func.my_string(u'hi', type_=Unicode) + ' ' +
-        ...       func.my_string(u'there', type_=Unicode))
-        my_string(:my_string_1) || :my_string_2 || my_string(:my_string_3)
-
-   The object returned by a :data:`.func` call is usually an instance of
-   :class:`.Function`.
-   This object meets the "column" interface, including comparison and labeling
-   functions.  The object can also be passed the :meth:`~.Connectable.execute`
-   method of a :class:`.Connection` or :class:`.Engine`, where it will be
-   wrapped inside of a SELECT statement first::
-
-        print(connection.execute(func.current_timestamp()).scalar())
-
-   In a few exception cases, the :data:`.func` accessor
-   will redirect a name to a built-in expression such as :func:`.cast`
-   or :func:`.extract`, as these names have well-known meaning
-   but are not exactly the same as "functions" from a SQLAlchemy
-   perspective.
-
-   .. versionadded:: 0.8 :data:`.func` can return non-function expression
-      constructs for common quasi-functional names like :func:`.cast`
-      and :func:`.extract`.
-
-   Functions which are interpreted as "generic" functions know how to
-   calculate their return type automatically. For a listing of known generic
-   functions, see :ref:`generic_functions`.
-
-   .. note::
-
-        The :data:`.func` construct has only limited support for calling
-        standalone "stored procedures", especially those with special
-        parameterization concerns.
-
-        See the section :ref:`stored_procedures` for details on how to use
-        the DBAPI-level ``callproc()`` method for fully traditional stored
-        procedures.
-
-"""
+func.__doc__ = _FunctionGenerator.__doc__
 
 modifier = _FunctionGenerator(group=False)
 
 
 class Function(FunctionElement):
-    """Describe a named SQL function.
+    r"""Describe a named SQL function.
+
+    The :class:`.Function` object is typically generated from the
+    :data:`.func` generation object.
+
+
+    :param \*clauses: list of column expressions that form the arguments
+     of the SQL function call.
+
+    :param type\_: optional :class:`.TypeEngine` datatype object that will be
+     used as the return value of the column expression generated by this
+     function call.
+
+    :param packagenames: a string which indicates package prefix names
+     to be prepended to the function name when the SQL is generated.
+     The :data:`.func` generator creates these when it is called using
+     dotted format, e.g.::
+
+        func.mypackage.some_function(col1, col2)
 
-    See the superclass :class:`.FunctionElement` for a description
-    of public methods.
 
     .. seealso::
 
+        :ref:`coretutorial_functions`
+
         :data:`.func` - namespace which produces registered or ad-hoc
         :class:`.Function` instances.
 
@@ -417,7 +641,7 @@ class Function(FunctionElement):
 
     """
 
-    __visit_name__ = 'function'
+    __visit_name__ = "function"
 
     def __init__(self, name, *clauses, **kw):
         """Construct a :class:`.Function`.
@@ -426,31 +650,44 @@ class Function(FunctionElement):
         new :class:`.Function` instances.
 
         """
-        self.packagenames = kw.pop('packagenames', None) or []
+        self.packagenames = kw.pop("packagenames", None) or []
         self.name = name
-        self._bind = kw.get('bind', None)
-        self.type = sqltypes.to_instance(kw.get('type_', None))
+        self._bind = kw.get("bind", None)
+        self.type = sqltypes.to_instance(kw.get("type_", None))
 
         FunctionElement.__init__(self, *clauses, **kw)
 
     def _bind_param(self, operator, obj, type_=None):
-        return BindParameter(self.name, obj,
-                             _compared_to_operator=operator,
-                             _compared_to_type=self.type,
-                             type_=type_,
-                             unique=True)
+        return BindParameter(
+            self.name,
+            obj,
+            _compared_to_operator=operator,
+            _compared_to_type=self.type,
+            type_=type_,
+            unique=True,
+        )
 
 
 class _GenericMeta(VisitableType):
     def __init__(cls, clsname, bases, clsdict):
         if annotation.Annotated not in cls.__mro__:
-            cls.name = name = clsdict.get('name', clsname)
-            cls.identifier = identifier = clsdict.get('identifier', name)
-            package = clsdict.pop('package', '_default')
+            cls.name = name = clsdict.get("name", clsname)
+            cls.identifier = identifier = clsdict.get("identifier", name)
+            package = clsdict.pop("package", "_default")
             # legacy
-            if '__return_type__' in clsdict:
-                cls.type = clsdict['__return_type__']
-            register_function(identifier, cls, package)
+            if "__return_type__" in clsdict:
+                cls.type = clsdict["__return_type__"]
+
+            # Check _register attribute status
+            cls._register = getattr(cls, "_register", True)
+
+            # Register the function if required
+            if cls._register:
+                register_function(identifier, cls, package)
+            else:
+                # Set _register to True to register child classes by default
+                cls._register = True
+
         super(_GenericMeta, cls).__init__(clsname, bases, clsdict)
 
 
@@ -512,33 +749,52 @@ class GenericFunction(util.with_metaclass(_GenericMeta, Function)):
 
     The above function will render as follows::
 
-        >>> print func.geo.buffer()
+        >>> print(func.geo.buffer())
         ST_Buffer()
 
-    .. versionadded:: 0.8 :class:`.GenericFunction` now supports
-       automatic registration of new functions as well as package
-       and custom naming support.
+    The name will be rendered as is, however without quoting unless the name
+    contains special characters that require quoting.  To force quoting
+    on or off for the name, use the :class:`.sqlalchemy.sql.quoted_name`
+    construct::
+
+        from sqlalchemy.sql import quoted_name
+
+        class GeoBuffer(GenericFunction):
+            type = Geometry
+            package = "geo"
+            name = quoted_name("ST_Buffer", True)
+            identifier = "buffer"
+
+    The above function will render as::
+
+        >>> print(func.geo.buffer())
+        "ST_Buffer"()
+
+    .. versionadded:: 1.3.13  The :class:`.quoted_name` construct is now
+       recognized for quoting when used with the "name" attribute of the
+       object, so that quoting can be forced on or off for the function
+       name.
 
-    .. versionchanged:: 0.8 The attribute name ``type`` is used
-       to specify the function's return type at the class level.
-       Previously, the name ``__return_type__`` was used.  This
-       name is still recognized for backwards-compatibility.
 
     """
 
     coerce_arguments = True
+    _register = False
 
     def __init__(self, *args, **kwargs):
-        parsed_args = kwargs.pop('_parsed_args', None)
+        parsed_args = kwargs.pop("_parsed_args", None)
         if parsed_args is None:
             parsed_args = [_literal_as_binds(c, self.name) for c in args]
+        self._has_args = self._has_args or bool(parsed_args)
         self.packagenames = []
-        self._bind = kwargs.get('bind', None)
+        self._bind = kwargs.get("bind", None)
         self.clause_expr = ClauseList(
-            operator=operators.comma_op,
-            group_contents=True, *parsed_args).self_group()
+            operator=operators.comma_op, group_contents=True, *parsed_args
+        ).self_group()
         self.type = sqltypes.to_instance(
-            kwargs.pop("type_", None) or getattr(self, 'type', None))
+            kwargs.pop("type_", None) or getattr(self, "type", None)
+        )
+
 
 register_function("cast", Cast)
 register_function("extract", Extract)
@@ -553,13 +809,15 @@ class next_value(GenericFunction):
     that does not provide support for sequences.
 
     """
+
     type = sqltypes.Integer()
     name = "next_value"
 
     def __init__(self, seq, **kw):
-        assert isinstance(seq, schema.Sequence), \
-            "next_value() accepts a Sequence object as input."
-        self._bind = kw.get('bind', None)
+        assert isinstance(
+            seq, schema.Sequence
+        ), "next_value() accepts a Sequence object as input."
+        self._bind = kw.get("bind", None)
         self.sequence = seq
 
     @property
@@ -568,8 +826,8 @@ class next_value(GenericFunction):
 
 
 class AnsiFunction(GenericFunction):
-    def __init__(self, **kwargs):
-        GenericFunction.__init__(self, **kwargs)
+    def __init__(self, *args, **kwargs):
+        GenericFunction.__init__(self, *args, **kwargs)
 
 
 class ReturnTypeFromArgs(GenericFunction):
@@ -577,28 +835,28 @@ class ReturnTypeFromArgs(GenericFunction):
 
     def __init__(self, *args, **kwargs):
         args = [_literal_as_binds(c, self.name) for c in args]
-        kwargs.setdefault('type_', _type_from_args(args))
-        kwargs['_parsed_args'] = args
+        kwargs.setdefault("type_", _type_from_args(args))
+        kwargs["_parsed_args"] = args
         super(ReturnTypeFromArgs, self).__init__(*args, **kwargs)
 
 
 class coalesce(ReturnTypeFromArgs):
+    _has_args = True
+
+
+class max(ReturnTypeFromArgs):  # noqa
     pass
 
 
-class max(ReturnTypeFromArgs):
+class min(ReturnTypeFromArgs):  # noqa
     pass
 
 
-class min(ReturnTypeFromArgs):
+class sum(ReturnTypeFromArgs):  # noqa
     pass
 
 
-class sum(ReturnTypeFromArgs):
-    pass
-
-
-class now(GenericFunction):
+class now(GenericFunction):  # noqa
     type = sqltypes.DateTime
 
 
@@ -614,19 +872,35 @@ class char_length(GenericFunction):
 
 
 class random(GenericFunction):
-    pass
+    _has_args = True
 
 
 class count(GenericFunction):
     r"""The ANSI COUNT aggregate function.  With no arguments,
     emits COUNT \*.
 
+    E.g.::
+
+        from sqlalchemy import func
+        from sqlalchemy import select
+        from sqlalchemy import table, column
+
+        my_table = table('some_table', column('id'))
+
+        stmt = select([func.count()]).select_from(my_table)
+
+    Executing ``stmt`` would emit::
+
+        SELECT count(*) AS count_1
+        FROM some_table
+
+
     """
     type = sqltypes.Integer
 
     def __init__(self, expression=None, **kwargs):
         if expression is None:
-            expression = literal_column('*')
+            expression = literal_column("*")
         super(count, self).__init__(expression, **kwargs)
 
 
@@ -681,7 +955,8 @@ class array_agg(GenericFunction):
     .. seealso::
 
         :func:`.postgresql.array_agg` - PostgreSQL-specific version that
-        returns :class:`.postgresql.ARRAY`, which has PG-specific operators added.
+        returns :class:`.postgresql.ARRAY`, which has PG-specific operators
+        added.
 
     """
 
@@ -689,8 +964,16 @@ class array_agg(GenericFunction):
 
     def __init__(self, *args, **kwargs):
         args = [_literal_as_binds(c) for c in args]
-        kwargs.setdefault('type_', self.type(_type_from_args(args)))
-        kwargs['_parsed_args'] = args
+
+        default_array_type = kwargs.pop("_default_array_type", sqltypes.ARRAY)
+        if "type_" not in kwargs:
+
+            type_from_args = _type_from_args(args)
+            if isinstance(type_from_args, sqltypes.ARRAY):
+                kwargs["type_"] = type_from_args
+            else:
+                kwargs["type_"] = default_array_type(type_from_args)
+        kwargs["_parsed_args"] = args
         super(array_agg, self).__init__(*args, **kwargs)
 
 
@@ -768,6 +1051,7 @@ class rank(GenericFunction):
     .. versionadded:: 1.1
 
     """
+
     type = sqltypes.Integer()
 
 
@@ -782,6 +1066,7 @@ class dense_rank(GenericFunction):
     .. versionadded:: 1.1
 
     """
+
     type = sqltypes.Integer()
 
 
@@ -796,6 +1081,7 @@ class percent_rank(GenericFunction):
     .. versionadded:: 1.1
 
     """
+
     type = sqltypes.Numeric()
 
 
@@ -810,4 +1096,70 @@ class cume_dist(GenericFunction):
     .. versionadded:: 1.1
 
     """
+
     type = sqltypes.Numeric()
+
+
+class cube(GenericFunction):
+    r"""Implement the ``CUBE`` grouping operation.
+
+    This function is used as part of the GROUP BY of a statement,
+    e.g. :meth:`.Select.group_by`::
+
+        stmt = select(
+            [func.sum(table.c.value), table.c.col_1, table.c.col_2]
+            ).group_by(func.cube(table.c.col_1, table.c.col_2))
+
+    .. versionadded:: 1.2
+
+    """
+    _has_args = True
+
+
+class rollup(GenericFunction):
+    r"""Implement the ``ROLLUP`` grouping operation.
+
+    This function is used as part of the GROUP BY of a statement,
+    e.g. :meth:`.Select.group_by`::
+
+        stmt = select(
+            [func.sum(table.c.value), table.c.col_1, table.c.col_2]
+        ).group_by(func.rollup(table.c.col_1, table.c.col_2))
+
+    .. versionadded:: 1.2
+
+    """
+    _has_args = True
+
+
+class grouping_sets(GenericFunction):
+    r"""Implement the ``GROUPING SETS`` grouping operation.
+
+    This function is used as part of the GROUP BY of a statement,
+    e.g. :meth:`.Select.group_by`::
+
+        stmt = select(
+            [func.sum(table.c.value), table.c.col_1, table.c.col_2]
+        ).group_by(func.grouping_sets(table.c.col_1, table.c.col_2))
+
+    In order to group by multiple sets, use the :func:`.tuple_` construct::
+
+        from sqlalchemy import tuple_
+
+        stmt = select(
+            [
+                func.sum(table.c.value),
+                table.c.col_1, table.c.col_2,
+                table.c.col_3]
+        ).group_by(
+            func.grouping_sets(
+                tuple_(table.c.col_1, table.c.col_2),
+                tuple_(table.c.value, table.c.col_3),
+            )
+        )
+
+
+    .. versionadded:: 1.2
+
+    """
+    _has_args = True
