@@ -1,5 +1,5 @@
 # util/_collections.py
-# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -50,7 +50,7 @@ class KeyedTuple(AbstractKeyedTuple):
         >>> k.two
         2
 
-    Result rows returned by :class:`_query.Query` that contain multiple
+    Result rows returned by :class:`.Query` that contain multiple
     ORM entities and/or column expressions make use of this
     class to return rows.
 
@@ -63,7 +63,7 @@ class KeyedTuple(AbstractKeyedTuple):
     receives its list of keys in place.   The subtype approach
     of ``collections.namedtuple()`` introduces significant complexity
     and performance overhead, which is not necessary for the
-    :class:`_query.Query` object's use case.
+    :class:`.Query` object's use case.
 
     .. seealso::
 
@@ -472,10 +472,13 @@ class IdentitySet(object):
 
     """
 
+    _working_set = set
+
     def __init__(self, iterable=None):
         self._members = dict()
         if iterable:
-            self.update(iterable)
+            for o in iterable:
+                self.add(o)
 
     def add(self, value):
         self._members[id(value)] = value
@@ -518,7 +521,7 @@ class IdentitySet(object):
             return True
 
     def issubset(self, iterable):
-        other = self.__class__(iterable)
+        other = type(self)(iterable)
 
         if len(self) > len(other):
             return False
@@ -539,7 +542,7 @@ class IdentitySet(object):
         return len(self) < len(other) and self.issubset(other)
 
     def issuperset(self, iterable):
-        other = self.__class__(iterable)
+        other = type(self)(iterable)
 
         if len(self) < len(other):
             return False
@@ -561,10 +564,11 @@ class IdentitySet(object):
         return len(self) > len(other) and self.issuperset(other)
 
     def union(self, iterable):
-        result = self.__class__()
-        members = self._members
-        result._members.update(members)
-        result._members.update((id(obj), obj) for obj in iterable)
+        result = type(self)()
+        # testlib.pragma exempt:__hash__
+        members = self._member_id_tuples()
+        other = _iter_id(iterable)
+        result._members.update(self._working_set(members).union(other))
         return result
 
     def __or__(self, other):
@@ -573,7 +577,7 @@ class IdentitySet(object):
         return self.union(other)
 
     def update(self, iterable):
-        self._members.update((id(obj), obj) for obj in iterable)
+        self._members = self.union(iterable)._members
 
     def __ior__(self, other):
         if not isinstance(other, IdentitySet):
@@ -582,12 +586,11 @@ class IdentitySet(object):
         return self
 
     def difference(self, iterable):
-        result = self.__class__()
-        members = self._members
-        other = {id(obj) for obj in iterable}
-        result._members.update(
-            ((k, v) for k, v in members.items() if k not in other)
-        )
+        result = type(self)()
+        # testlib.pragma exempt:__hash__
+        members = self._member_id_tuples()
+        other = _iter_id(iterable)
+        result._members.update(self._working_set(members).difference(other))
         return result
 
     def __sub__(self, other):
@@ -605,12 +608,11 @@ class IdentitySet(object):
         return self
 
     def intersection(self, iterable):
-        result = self.__class__()
-        members = self._members
-        other = {id(obj) for obj in iterable}
-        result._members.update(
-            (k, v) for k, v in members.items() if k in other
-        )
+        result = type(self)()
+        # testlib.pragma exempt:__hash__
+        members = self._member_id_tuples()
+        other = _iter_id(iterable)
+        result._members.update(self._working_set(members).intersection(other))
         return result
 
     def __and__(self, other):
@@ -628,17 +630,18 @@ class IdentitySet(object):
         return self
 
     def symmetric_difference(self, iterable):
-        result = self.__class__()
-        members = self._members
-        other = {id(obj): obj for obj in iterable}
+        result = type(self)()
+        # testlib.pragma exempt:__hash__
+        members = self._member_id_tuples()
+        other = _iter_id(iterable)
         result._members.update(
-            ((k, v) for k, v in members.items() if k not in other)
-        )
-        result._members.update(
-            ((k, v) for k, v in other.items() if k not in members)
+            self._working_set(members).symmetric_difference(other)
         )
         return result
 
+    def _member_id_tuples(self):
+        return ((id(v), v) for v in self._members.values())
+
     def __xor__(self, other):
         if not isinstance(other, IdentitySet):
             return NotImplemented
@@ -706,6 +709,13 @@ class WeakSequence(object):
 
 
 class OrderedIdentitySet(IdentitySet):
+    class _working_set(OrderedSet):
+        # a testing pragma: exempt the OIDS working set from the test suite's
+        # "never call the user's __hash__" assertions.  this is a big hammer,
+        # but it's safe here: IDS operates on (id, instance) tuples in the
+        # working set.
+        __sa_hash_exempt__ = True
+
     def __init__(self, iterable=None):
         IdentitySet.__init__(self)
         self._members = OrderedDict()
@@ -1070,6 +1080,13 @@ class ThreadLocalRegistry(ScopedRegistry):
             pass
 
 
+def _iter_id(iterable):
+    """Generator: ((id(o), o) for o in iterable)."""
+
+    for item in iterable:
+        yield id(item), item
+
+
 def has_dupes(sequence, target):
     """Given a sequence and search object, return True if there's more
     than one, False if zero or one of them.
